#### The code to post on website

####### Load Libraries to be used ##############
library(CollocInfer)
library(deSolve)
library(lokern)
library(MASS)

####### Set seed so that get same results when rerun #######
set.seed(123457)
####### Set kernel K(u)=(1-u^2)_{+}
kernel.e<-function(x){
    3/4*(1-x^2)*(1>x^2)+1e-10
}

###### Input data ##############
###### Here assume the data is in the two csv files 
######   in the same directory where R started.
######   May need to adjust location on your system.
workdata<-read.csv("./data_cd8.csv",
             header = TRUE, sep = ",", quote="\"", dec=".",
             fill = TRUE, comment.char="")
workdataD<-read.csv("./data_dc.csv",
              header = TRUE, sep = ",", quote="\"", dec=".",
              fill = TRUE, comment.char="")

###### Clean data: drop the "NA" case and fit ODE from day 5 to day 14 
workdata<-data.frame(workdata)
workdata<-workdata[!is.na(workdata$data_lun_2008),]
fitteddata<-workdata[workdata$time>=5 & workdata$time<=14,]


###### Use spline to smooth covariate Dt for use in ODE
dt.spline<-smooth.spline(workdataD$TIME,workdataD$LOGDC,all.knots=T)
###### Define the ODE function (on log scale) ###########
ODEmod <- function(Time, State, Pars) {
  with(as.list(c(State, Pars)), {
    Dt<-10^glkerns(workdataD$TIME,workdataD$LOGDC,x.out=Time+5-3.08,deriv=0)$est
    dX1 <- (rm*Dt-dm-gms-gml)                 #log(Tm)
    dX2 <- (rs*Dt-ds-gsl) + gms*exp(X1-X2)    #log(Ts)
    dX3 <- gml*exp(X1-X3)+gsl*exp(X2-X3) - dl #log(Tl)

    return(list(c(dX1, dX2, dX3)))
  })
}
##### Miao etal (2011)'s fit
pars    <- c(rm=1.66e-5,rs=4.483e-5,
             dm=0,ds=0,dl=3.96,
             gms=0.157,gml=0,gsl=0.495)
xini<-c(X1=log(3.96e+3),X2=log(3.64e+4),X3=log(1.31e+3))
##### initial states from data instead
xini0<-mean(log(fitteddata[fitteddata$time==5,]))[2:4]
names(xini0)=names(xini)

##### Compute RSS (log10 scale) for Miao Etal's fit
out     <- as.data.frame(ode(func = ODEmod, y = xini,
                         parms = pars, times = fitteddata$time-5))
sum(((out[,2:4]-log(fitteddata[2:4]))/log(10))^2)
#[1] 15.77025

########################################################
####### Do NLS #########################################
########################################################
yfit<-log(c(fitteddata$data_mln_2008,fitteddata$data_spl_2008,fitteddata$data_lun_2008))
ODEmod.0dsdmgml <- function(Time, State, Pars) {
  with(as.list(c(State, Pars)), {
    Dt<-10^predict(dt.spline,Time+5-3.08)$y
    dX1 <- (rm*Dt-gms)                 #log(Tm)
    dX2 <- (rs*Dt-gsl) + gms*exp(X1-X2)    #log(Ts)
    dX3 <- gsl*exp(X2-X3) - dl #log(Tl)

    return(list(c(dX1, dX2, dX3)))
  })
}
## To calculate things quicker, 
## get the time points and repeated frequency to avoid ODE overlap
time.out<-fitteddata$time-5
time.0<-unique(time.out)
time.freq<-time.0-time.0
for (i in seq(length(time.freq))){
   time.freq[i]<-sum(time.out==time.0[i])}
#rep(time.0,time.freq)-time.out
ODEsolver.0dsdmgml<-function(X1,X2,X3, theparms) {
   names(X1)<-"X1"
   names(X2)<-"X2"
   names(X3)<-"X3"
   names(theparms)<-c("rm","rs","dl",
             "gms","gsl")
   tmp<-ode(func = ODEmod.0dsdmgml, y = c(X1,X2,X3),parms = theparms, times = time.0)
   tmp<-apply(tmp,2,function(x) rep(x,time.freq))
   return(c(tmp[,"X1"],tmp[,"X2"],tmp[,"X3"]))
}

#######################
### Look at a grid of starting values
### each one from 0,0.5,1
######################
k.g<-3
niter<-4
conv.case<-array(NA,rep(k.g,5))
conv.time<-array(NA,rep(k.g,5))
conv1<-NULL
time1<-NULL
para.start<-NULL
para.est<-NULL
para.est1<-NULL
rss<-NULL
para.start<-NULL
now=proc.time()
for (k1 in seq(k.g)) {
  for (k2 in seq(k.g)) {
    for (k3 in seq(k.g)) {
      for (k4 in seq(k.g)) {
        for (k5 in seq(k.g)) {
          now1=proc.time()
          tmp=try(nls(yfit ~ ODEsolver.0dsdmgml(x1,x2,x3, thepar), 
                  control=nls.control(warnOnly=T,tol=1e-4),
                  start=list(x1=xini[1],x2=xini[2],x3=xini[3], 
                     thepar=(c(k1,k2,k3,k4,k5)-1)/(k.g-1))))
          conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
          if (class(tmp)=="try-error") {
            conv.case[k1,k2,k3,k4,k5]<-0
          }
          else {
            if (min(coef(tmp)[-seq(3)])<0){
              conv.case[k1,k2,k3,k4,k5]<-(-1)
            }
            else {  
              conv.case[k1,k2,k3,k4,k5]<-1
            } 
            para.est1<-rbind(para.est1,coef(tmp))
            rss<-rbind(rss,sum((resid(tmp)/log(10))^2))
            conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
            time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
            para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5)-1)/(k.g-1))
            }
          }
}}}}
usedtime=proc.time()-now
#### Summarize results
write.table(cbind(rss,time1,conv1,para.est1,para.start),file="resultsNLS.txt")
mean(conv.time)
#[1] 13.18226
min(rss)
#[1] 15.76867  ### About the same as Miao etal's

############################################
###### PsLS ################################
############################################
n<-length(workdata$time)
nt<-21      ## m, number of time points we did the smoothing
tobs<-workdata$time
y1<-log(workdata$data_mln_2008)
y2<-log(workdata$data_spl_2008)
y3<-log(workdata$data_lun_2008)
y<-rbind(y1,y2,y3) # d by n matrix
## The time grid is from day 5 to day 14
## The time grid for DC with time delay of 3.08
tt<-5+(seq(nt)-1)/(nt-1)*(14-5)
tt.d<-tt-3.08

##### Fit the covariate D at delayed time t-3.08
logDt<-predict(dt.spline,tt.d)$y
Dt<-10^logDt
##### Fit the derivative of covariate D at delayed time t-3.08
logDt.der<-predict(dt.spline,tt.d,deriv=1)$y
Dt.der<-log(10)*Dt*logDt.der

##### The observed times grid, for computation purpose later
d=3                                 # number of variables
p=length(pars)                      # number of parameters
ttMat<-matrix(rep(tobs,rep(d,n)),nrow=d) # d by n matrix

### For weighted version, delete the two boundary points (weight=0)
nt.w<-nt-2
tt.w<-tt[seq(nt.w)+1]
## Defines weights used in fitting later: 
##   piecewise linear weight going to zero at boundary
W1<-tt/tt
W1[tt<6]<-(tt[tt<6]-5)
W1[tt>13]<-14-tt[tt>13]
W1.w<-W1[seq(nt.w)+1]

#### Creat the array to store t_i - t_{k}^* for i=1,...,n, k=1,...,nt
#### So we have d by n by nt (m in paper) array
tDiff<-array(ttMat,c(d,n,nt))-aperm(array(rep(tt,rep(d,nt)),c(d,nt,n)),c(1,3,2))
#### delete two boundary points
tDiff.w<-array(ttMat,c(d,n,nt.w))-aperm(array(rep(tt.w,rep(d,nt.w)),c(d,nt.w,n)),c(1,3,2))

######## Data augumentation. Increase to m points of fits
###### Use glkerns to estimate X and derivative of X over 
######     a grid time points
## the bandwidth selected automatically using data from 4 to 14 days
##  not the whole data
#> lok.fit1$bandwidth
#[1] 3.483761
#> lok.fit2$bandwidth
#[1] 0.8498683
#> lok.fit3$bandwidth
#[1] 1.390954
#> lok.derfit1$bandwidth
#[1] 2.064513
#> lok.derfit2$bandwidth
#[1] 1.902007
#> lok.derfit3$bandwidth
#[1] 1.387027
  lok.fit1<-glkerns(tobs,y1,x.out=tt,deriv=0,bandwidth=3.483761)
  lok.fit2<-glkerns(tobs,y2,x.out=tt,deriv=0,bandwidth=0.8498683)
  lok.fit3<-glkerns(tobs,y3,x.out=tt,deriv=0,bandwidth=1.390954)
  lok.derfit1<-glkerns(tobs,y1,x.out=tt,deriv=1,bandwidth=2.064513)
  lok.derfit2<-glkerns(tobs,y2,x.out=tt,deriv=1,bandwidth=1.902007)
  lok.derfit3<-glkerns(tobs,y3,x.out=tt,deriv=1,bandwidth=1.387027)

  h.glkerns<-rbind(rep(lok.fit1$bandwidth,nt),rep(lok.fit2$bandwidth,nt),rep(lok.fit3$bandwidth,nt))
  h.glkerns.der<-rbind(rep(lok.derfit1$bandwidth,nt),rep(lok.derfit2$bandwidth,nt),rep(lok.derfit3$bandwidth))

## calcuate the variance using repeat measurements
tobs.real<-unique(tobs)
y1mean<-y1
y2mean<-y2
y3mean<-y3
for (i in tobs.real) {
   y1mean[tobs==i]=mean(y1mean[tobs==i])
   y2mean[tobs==i]=mean(y2mean[tobs==i])
   y3mean[tobs==i]=mean(y3mean[tobs==i])
}
SigmaSq1<-sum((y1-y1mean)^2)/(length(tobs)-length(tobs.real))
SigmaSq2<-sum((y2-y2mean)^2)/(length(tobs)-length(tobs.real))
SigmaSq3<-sum((y3-y3mean)^2)/(length(tobs)-length(tobs.real))
#> SigmaSq1
#[1] 0.4687576
#> SigmaSq2
#[1] 0.08662212
#> SigmaSq3
#[1] 0.3147943

now=proc.time()
  ### PsLs as in Liang-Wu
  h.fac<-n^(-3/35)*(log(n))^(-1/16)
  ##### Estimate X using h with local linear fit
  h<-h.glkerns*h.fac
  KY1<-KY0<-K0<-array(dim=c(d,n,nt))   ## kernel weights 
  for (k in 1:nt){
    K0[,,k]<-kernel.e(tDiff[,,k]/h[,k])/h[,k]
  }
    K1<-K0*tDiff
    K2<-K0*tDiff^2
  for (k in 1:nt){
    KY0[,,k]<-K0[,,k]*y
    KY1[,,k]<-K1[,,k]*y
  }
  S0<-apply(K0,c(1,3),sum)
  S1<-apply(K1,c(1,3),sum)
  S2<-apply(K2,c(1,3),sum)
  SY0<-apply(KY0,c(1,3),sum)
  SY1<-apply(KY1,c(1,3),sum)

  hatX<-matrix(nrow=2*d,ncol=nt)
  for (i in 1:nt) {
    S<-rbind(cbind(diag(S0[,i]+1e-10),diag(S1[,i])),
           cbind(diag(S1[,i]),diag(S2[,i]+1e-10)))
  hatX[,i]<-solve(S,c(SY0[,i],SY1[,i]))
  }
  Xfit<-hatX[1:d,]

 ##### Estimate derivative using h.der with local quadratic fit
  h<-h.glkerns.der*h.fac
  KY2<-KY1<-KY0<-K0<-array(dim=c(d,n,nt))   ## kernel weights 
  for (k in 1:nt){
    K0[,,k]<-kernel.e(tDiff[,,k]/h[,k])/h[,k]
  }
    K1<-K0*tDiff
    K2<-K0*tDiff^2
    K3<-K0*tDiff^3
    K4<-K0*tDiff^4
  for (k in 1:nt){
    KY0[,,k]<-K0[,,k]*y
    KY1[,,k]<-K1[,,k]*y
    KY2[,,k]<-K2[,,k]*y
  }
  S0<-apply(K0,c(1,3),sum)
  S1<-apply(K1,c(1,3),sum)
  S2<-apply(K2,c(1,3),sum)
  S3<-apply(K3,c(1,3),sum)
  S4<-apply(K4,c(1,3),sum)
  SY0<-apply(KY0,c(1,3),sum)
  SY1<-apply(KY1,c(1,3),sum)
  SY2<-apply(KY2,c(1,3),sum)

  hatX<-matrix(nrow=3*d,ncol=nt)
  for (i in 1:nt) {
  S<-rbind(cbind(diag(S0[,i]+1e-10),diag(S1[,i]),diag(S2[,i])),
           cbind(diag(S1[,i]),diag(S2[,i]+1e-10),diag(S3[,i])),
           cbind(diag(S2[,i]),diag(S3[,i]),diag(S4[,i]+1e-10)))
  hatX[,i]<-solve(S,c(SY0[,i],SY1[,i],SY2[,i]))
  }

  derfit<-hatX[(d+1):(2*d),]

  ### Do PsLS: 
  ### Notice that we stack responses X1', X2' and X3' together 
  ###   as one response variable.
  hatDerivative<-c(derfit[1,],derfit[2,],derfit[3,])
  hatX1<-Xfit[1,]
  hatX2<-Xfit[2,]
  hatX3<-Xfit[3,]

now=proc.time()
   ### Drop the 0 coefficients according to Wu's paper
    nlsFit1.a.0<-nls(hatDerivative ~ c(exp(rm)*Dt-exp(gms),(exp(rs)*Dt-exp(gsl)) + exp(gms)*exp(hatX1-hatX2), exp(gsl)*exp(hatX2-hatX3) - exp(dl)),
        control=nls.control(warnOnly=T),
        start=log(pars0+1e-7),
        weights=c(W1,W1,W1))
     exp(coef(nlsFit1.a.0))
usedtime=proc.time()-now
usedtime
#   user  system elapsed 
#   0.03    0.30    0.32
#### PsLS just did once. Did not try the grid of starting values of 
####     0, 0.5, 1 as other methods. Since this starting values on log 
####     scale, can not use 0. If do not do log on parameters as other 
####     methods, then always get negative parameters. 
####     PsLS does not have convergent problem. Can use almost any 
####     starting values. Its solution are not reasonable, however.
####     (Extremely large RSS).

############################################################################## OCLP estimator ################################################
#######################################################################
  h<-h.glkerns.der[,seq(nt.w)+1]*h.fac
  KY2<-KY1<-KY0<-K0<-array(dim=c(d,n,nt.w))   ## kernel weights 
  W<-rbind(W1.w,W1.w,W1.w)  
  for (k in 1:nt.w){
    K0[,,k]<-kernel.e(tDiff.w[,,k]/h[,k])/h[,k]*W[,k]
  }
    K1<-K0*tDiff.w
    K2<-K0*tDiff.w^2
    K3<-K0*tDiff.w^3
    K4<-K0*tDiff.w^4
  for (k in 1:nt.w){
    KY0[,,k]<-K0[,,k]*y
    KY1[,,k]<-K1[,,k]*y
    KY2[,,k]<-K2[,,k]*y
  }
  S0<-apply(K0,c(1,3),sum)
  S1<-apply(K1,c(1,3),sum)
  S2<-apply(K2,c(1,3),sum)
  S3<-apply(K3,c(1,3),sum)
  S4<-apply(K4,c(1,3),sum)
  SY0<-apply(KY0,c(1,3),sum)
  SY1<-apply(KY1,c(1,3),sum)
  SY2<-apply(KY2,c(1,3),sum)

  X1.last<-lok.fit1$est[seq(nt.w)+1]
  X2.last<-lok.fit2$est[seq(nt.w)+1]
  X3.last<-lok.fit3$est[seq(nt.w)+1]
  X.last<-rbind(X1.last,X2.last,X3.last) 
  # use X from fits by glkerns instead of LW


Z<-Dt[seq(nt.w)+1]
Zder<-Dt.der[seq(nt.w)+1]

#########################################
#### Define one iteration as a function
#########################################
  oneiteration.dm.ds.gml<-function(X,THETA,Z,Zder,...){
    d<-dim(X)[1]
    p<-length(THETA)
    nt.w<-dim(X)[2]

    X1<-X[1,]
    X2<-X[2,]
    X3<-X[3,]

    rhom<-THETA["rm"]
    rhos<-THETA["rs"]
    dm<-THETA["dm"]
    ds<-THETA["ds"]
    dl<-THETA["dl"]
    gms<-THETA["gms"]
    gml<-THETA["gml"]
    gsl<-THETA["gsl"]

    f<-array(dim=c(d,1,nt.w))    #F
    f[1,1,]<- (rhom*Z-gms)
    f[2,1,]<- (rhos*Z-gsl) + gms*exp(X1-X2)
    f[3,1,]<- gsl*exp(X2-X3) - dl 

    f1<-array(dim=c(d,1,nt.w))   #dF/dt
    f1[1,1,]<- rhom*Zder
    f1[2,1,]<- rhos*Zder + gms*exp(X1-X2)*(f[1,1,]-f[2,1,])
    f1[3,1,]<- gsl*exp(X2-X3)*(f[2,1,]-f[3,1,])

    Dalpha<-array(dim=c(d,d,nt.w))  #dF/Dx
    Dalpha[1,,]<- 0
    Dalpha[2,1,]<- gms*exp(X1-X2)
    Dalpha[2,2,]<- (-gms*exp(X1-X2))
    Dalpha[2,3,]<- 0
    Dalpha[3,1,]<- 0
    Dalpha[3,2,]<- gsl*exp(X2-X3) 
    Dalpha[3,3,]<- (-gsl*exp(X2-X3))

    Dalpha1<-array(dim=c(d,d,nt.w)) #d(dF/dt)/dx
    Dalpha1[1,,]<- 0
    Dalpha1[2,1,]<- gms*exp(X1-X2)*(f[1,1,]-f[2,1,]+Dalpha[1,1,]-Dalpha[2,1,])
    Dalpha1[2,2,]<- gms*exp(X1-X2)*(-f[1,1,]+f[2,1,]+Dalpha[1,2,]-Dalpha[2,2,])
    Dalpha1[2,3,]<- 0
    Dalpha1[3,1,]<- gsl*exp(X2-X3)*(Dalpha[2,1,]-Dalpha[3,1,])
    Dalpha1[3,2,]<- gsl*exp(X2-X3)*(f[2,1,]-f[3,1,]+Dalpha[2,2,]-Dalpha[3,2,])
    Dalpha1[3,3,]<- gsl*exp(X2-X3)*(-f[2,1,]+f[3,1,]+Dalpha[2,3,]-Dalpha[3,3,])

    tildeDalpha<-array(dim=c(d,d,n,nt.w))
    tildeDalpha[1,1,,]<-1+t(t(tDiff.w[1,,])*Dalpha[1,1,])+t(t(tDiff.w[1,,]^2)*Dalpha1[1,1,])/2
    tildeDalpha[1,2,,]<-t(t(tDiff.w[1,,])*Dalpha[1,2,])+t(t(tDiff.w[1,,]^2)*Dalpha1[1,2,])/2
    tildeDalpha[1,3,,]<-t(t(tDiff.w[1,,])*Dalpha[1,3,])+t(t(tDiff.w[1,,]^2)*Dalpha1[1,3,])/2
    tildeDalpha[2,1,,]<-t(t(tDiff.w[2,,])*Dalpha[2,1,])+t(t(tDiff.w[2,,]^2)*Dalpha1[2,1,])/2
    tildeDalpha[2,2,,]<-1+t(t(tDiff.w[2,,])*Dalpha[2,2,])+t(t(tDiff.w[2,,]^2)*Dalpha1[2,2,])/2
    tildeDalpha[2,3,,]<-t(t(tDiff.w[2,,])*Dalpha[2,3,])+t(t(tDiff.w[2,,]^2)*Dalpha1[2,3,])/2
    tildeDalpha[3,1,,]<-t(t(tDiff.w[3,,])*Dalpha[3,1,])+t(t(tDiff.w[3,,]^2)*Dalpha1[3,1,])/2
    tildeDalpha[3,2,,]<-t(t(tDiff.w[3,,])*Dalpha[3,2,])+t(t(tDiff.w[3,,]^2)*Dalpha1[3,2,])/2
    tildeDalpha[3,3,,]<-1+t(t(tDiff.w[3,,])*Dalpha[3,3,])+t(t(tDiff.w[3,,]^2)*Dalpha1[3,3,])/2

    tildeDalphaSq<-array(dim=c(d,d,n,nt.w))
    tildeDalphaSq[1,1,,]<-tildeDalpha[1,1,,]^2*K0[1,,] +tildeDalpha[2,1,,]^2*K0[2,,]+tildeDalpha[3,1,,]^2*K0[3,,]
    tildeDalphaSq[2,2,,]<-tildeDalpha[1,2,,]^2*K0[1,,] +tildeDalpha[2,2,,]^2*K0[2,,]+tildeDalpha[3,2,,]^2*K0[3,,]
    tildeDalphaSq[3,3,,]<-tildeDalpha[1,3,,]^2*K0[1,,] +tildeDalpha[2,3,,]^2*K0[2,,]+tildeDalpha[3,3,,]^2*K0[3,,]
    tildeDalphaSq[2,1,,]<-tildeDalphaSq[1,2,,]<-tildeDalpha[1,1,,]*tildeDalpha[1,2,,]*K0[1,,] +tildeDalpha[2,1,,]*tildeDalpha[2,2,,]*K0[2,,] +tildeDalpha[3,1,,]*tildeDalpha[3,2,,]*K0[3,,]
    tildeDalphaSq[3,1,,]<-tildeDalphaSq[1,3,,]<-tildeDalpha[1,1,,]*tildeDalpha[1,3,,]*K0[1,,] +tildeDalpha[2,1,,]*tildeDalpha[2,3,,]*K0[2,,] +tildeDalpha[3,1,,]*tildeDalpha[3,3,,]*K0[3,,]
    tildeDalphaSq[3,2,,]<-tildeDalphaSq[2,3,,]<-tildeDalpha[1,2,,]*tildeDalpha[1,3,,]*K0[1,,] +tildeDalpha[2,2,,]*tildeDalpha[2,3,,]*K0[2,,] +tildeDalpha[3,2,,]*tildeDalpha[3,3,,]*K0[3,,]

    Dtheta<-array(dim=c(d,p,nt.w)) #dF/Dtheta
    Dtheta[1,1,]<-Z
    Dtheta[1,2,]<-Dtheta[1,4:5,]<-Dtheta[1,8,]<-0
    Dtheta[1,3,]<- 0
    Dtheta[1,6,]<- (-1)
    Dtheta[1,7,]<- 0
    Dtheta[2,1,]<- Dtheta[2,3,]<- Dtheta[2,5,]<- Dtheta[2,7,]<- 0
    Dtheta[2,2,]<- Z
    Dtheta[2,4,]<- 0
    Dtheta[2,6,]<- exp(X1-X2)
    Dtheta[2,8,]<- (-1)
    Dtheta[3,1:4,]<- Dtheta[3,6,]<- 0
    Dtheta[3,5,]<- (-1)
    Dtheta[3,7,]<- 0
    Dtheta[3,8,]<- exp(X2-X3) 

    Dtheta1<-array(dim=c(d,p,nt.w))
    Dtheta1[1,1,]<-Zder
    Dtheta1[1,2:8,]<- 0
    Dtheta1[2,,]<- gms*exp(X1-X2)*(Dtheta[1,,]-Dtheta[2,,])
    Dtheta1[2,2,]<- Dtheta1[2,2,]+Zder 
    Dtheta1[2,6,]<- Dtheta1[2,6,]+exp(X1-X2)*(f[1,1,]-f[2,1,])
    Dtheta1[3,,]<- gsl*exp(X2-X3)*(Dtheta[2,,]-Dtheta[3,,])
    Dtheta1[3,7,]<- 0
    Dtheta1[3,8,]<- Dtheta1[3,8,]+exp(X2-X3)*(f[2,1,]-f[3,1,])

    tildeDtheta<-array(dim=c(d,p,n,nt.w))
    for (k in 1:p) {
      tildeDtheta[1,k,,]<-t(t(tDiff.w[1,,])*Dtheta[1,k,])+t(t(tDiff.w[1,,]^2)*Dtheta1[1,k,])/2
      tildeDtheta[2,k,,]<-t(t(tDiff.w[2,,])*Dtheta[2,k,])+t(t(tDiff.w[2,,]^2)*Dtheta1[2,k,])/2
      tildeDtheta[3,k,,]<-t(t(tDiff.w[3,,])*Dtheta[3,k,])+t(t(tDiff.w[3,,]^2)*Dtheta1[3,k,])/2
    }

    tildeDthetaSq<-array(dim=c(p,p,n,nt.w))
    for (k1 in 1:p) {
      for (k2 in 1:p) {
        tildeDthetaSq[k1,k2,,]<-tildeDtheta[1,k1,,]*tildeDtheta[1,k2,,]*K0[1,,] +tildeDtheta[2,k1,,]*tildeDtheta[2,k2,,]*K0[2,,]+tildeDtheta[3,k1,,]*tildeDtheta[3,k2,,]*K0[3,,]
      }
    }

    tildeDalphaTheta<-array(dim=c(d,p,n,nt.w))
    for (k in 1:p) {
      tildeDalphaTheta[1,k,,]<-tildeDalpha[1,1,,]*tildeDtheta[1,k,,]*K0[1,,]+tildeDalpha[2,1,,]*tildeDtheta[2,k,,]*K0[2,,]+tildeDalpha[3,1,,]*tildeDtheta[3,k,,]*K0[3,,]
      tildeDalphaTheta[2,k,,]<-tildeDalpha[1,2,,]*tildeDtheta[1,k,,]*K0[1,,]+tildeDalpha[2,2,,]*tildeDtheta[2,k,,]*K0[2,,]+tildeDalpha[3,2,,]*tildeDtheta[3,k,,]*K0[3,,]
      tildeDalphaTheta[3,k,,]<-tildeDalpha[1,3,,]*tildeDtheta[1,k,,]*K0[1,,]+tildeDalpha[2,3,,]*tildeDtheta[2,k,,]*K0[2,,]+tildeDalpha[3,3,,]*tildeDtheta[3,k,,]*K0[3,,]
    }

    fMinus<-array(dim=c(d,1,nt.w))
    for (k in 1:nt.w){
      fMinus[,,k]<-f[,,k]-Dtheta[,,k]%*%THETA-Dalpha[,,k]%*%X[,k]
    }

    fMinus1<-array(dim=c(d,1,nt.w))
    for (k in 1:nt.w){
      fMinus1[,,k]<-f1[,,k]-Dtheta1[,,k]%*%THETA-Dalpha1[,,k]%*%X[,k]
    }

    Yall<-array(y,dim=c(d,n,nt.w))
    tildeY<-array(dim=c(d,n,nt.w))
    tildeY[1,,]<-Yall[1,,]-t(t(tDiff.w[1,,])*fMinus[1,1,])-t(t(tDiff.w[1,,]^2)*fMinus1[1,1,])/2
    tildeY[2,,]<-Yall[2,,]-t(t(tDiff.w[2,,])*fMinus[2,1,])-t(t(tDiff.w[2,,]^2)*fMinus1[2,1,])/2
    tildeY[3,,]<-Yall[3,,]-t(t(tDiff.w[3,,])*fMinus[3,1,])-t(t(tDiff.w[3,,]^2)*fMinus1[3,1,])/2

    tildeDalphaY<-array(dim=c(d,n,nt.w))
    tildeDalphaY[1,,]<-tildeDalpha[1,1,,]*tildeY[1,,]*K0[1,,]+tildeDalpha[2,1,,]*tildeY[2,,]*K0[2,,]+tildeDalpha[3,1,,]*tildeY[3,,]*K0[3,,]
    tildeDalphaY[2,,]<-tildeDalpha[1,2,,]*tildeY[1,,]*K0[1,,]+tildeDalpha[2,2,,]*tildeY[2,,]*K0[2,,]+tildeDalpha[3,2,,]*tildeY[3,,]*K0[3,,]
    tildeDalphaY[3,,]<-tildeDalpha[1,3,,]*tildeY[1,,]*K0[1,,]+tildeDalpha[2,3,,]*tildeY[2,,]*K0[2,,]+tildeDalpha[3,3,,]*tildeY[3,,]*K0[3,,]

    tildeDthetaY<-array(dim=c(p,n,nt.w))
    for (k in 1:p){
      tildeDthetaY[k,,]<-tildeDtheta[1,k,,]*tildeY[1,,]*K0[1,,]+tildeDtheta[2,k,,]*tildeY[2,,]*K0[2,,]+tildeDtheta[3,k,,]*tildeY[3,,]*K0[3,,]
    }

    tildeDalphaSq<-array(dim=c(d,d,n,nt.w))
    tildeDalphaSq[1,1,,]<-tildeDalpha[1,1,,]^2*K0[1,,]+tildeDalpha[2,1,,]^2*K0[2,,]+tildeDalpha[3,1,,]^2*K0[3,,]
    tildeDalphaSq[2,2,,]<-tildeDalpha[1,2,,]^2*K0[1,,]+tildeDalpha[2,2,,]^2*K0[2,,]+tildeDalpha[3,2,,]^2*K0[3,,]
    tildeDalphaSq[3,3,,]<-tildeDalpha[1,3,,]^2*K0[1,,]+tildeDalpha[2,3,,]^2*K0[2,,]+tildeDalpha[3,3,,]^2*K0[3,,]
    tildeDalphaSq[2,1,,]<-tildeDalphaSq[1,2,,]<-tildeDalpha[1,1,,]*tildeDalpha[1,2,,]*K0[1,,]+tildeDalpha[2,1,,]*tildeDalpha[2,2,,]*K0[2,,]+tildeDalpha[3,1,,]*tildeDalpha[3,2,,]*K0[3,,]
    tildeDalphaSq[3,1,,]<-tildeDalphaSq[1,3,,]<-tildeDalpha[1,1,,]*tildeDalpha[1,3,,]*K0[1,,]+tildeDalpha[2,1,,]*tildeDalpha[2,3,,]*K0[2,,]+tildeDalpha[3,1,,]*tildeDalpha[3,3,,]*K0[3,,]
    tildeDalphaSq[2,3,,]<-tildeDalphaSq[3,2,,]<-tildeDalpha[1,3,,]*tildeDalpha[1,2,,]*K0[1,,]+tildeDalpha[2,3,,]*tildeDalpha[2,2,,]*K0[2,,]+tildeDalpha[3,3,,]*tildeDalpha[3,2,,]*K0[3,,]

    sumDalphaSq<-apply(tildeDalphaSq,c(1,2,4),sum)
    sumDalphaTheta<-apply(tildeDalphaTheta,c(1,2,4),sum)
    sumDalphaY<-apply(tildeDalphaY,c(1,3),sum)

    #### Compute (T^tWT)^{-1}(T^tWY) 
    tmpDiag<-matrix(0,nrow=d*nt.w,ncol=d*nt.w)

    for (k in 1:nt.w){tmpDiag[(k-1)*d+(1:d),(k-1)*d+(1:d)]<-sumDalphaSq[,,k]} 

    tmpLastCol<-matrix(0,nrow=d*nt.w,ncol=p)
    for (k in 1:nt.w) {tmpLastCol[(k-1)*d+(1:d),]<-sumDalphaTheta[,,k]}

    tmpLast<-apply(tildeDthetaSq,c(1,2),sum)
    S<-rbind(cbind(tmpDiag,tmpLastCol),cbind(t(tmpLastCol),tmpLast))

    SY<-vector(length=d*nt.w+p)
    for (k in 1:nt.w) {SY[(k-1)*d+(1:d)]<-sumDalphaY[,k]}
    SY[d*nt.w+(1:p)]<-apply(tildeDthetaY,1,sum)

    #To get rid of singular matrix
    S[d*nt.w+(1:p),d*nt.w+(1:p)]<-S[d*nt.w+(1:p),d*nt.w+(1:p)]+diag(1e-8,p)
    ### setting zero rates create singular matrix,
    ### drop the rows and columns corresponding to those rates
    idkeep<-seq(d*nt.w+p)[-(d*nt.w+c(3,4,7))]
    Skeep<-S[idkeep,idkeep]
    SYkeep<-SY[idkeep]
    newfit<-solve(Skeep,SYkeep)
    newX<-matrix(newfit[seq(d*nt.w)],nrow=d)
    newTHETA<-newfit[d*nt.w+(1:(p-3))]
    names(newTHETA)<-names(THETA[-c(3,4,7)])

    WSSE<-sum(tildeY^2*K0)-sum(newfit*SYkeep)

    return(list(newX=newX,newTHETA=newTHETA,WSSE=WSSE))
  }

#######################
### Look at a grid of starting values
### each one from 0,0.5,1
######################
k.g<-3
niter<-10
conv.case<-array(NA,rep(k.g,5))
conv.time<-array(NA,rep(k.g,5))
conv1<-NULL
time1<-NULL
para.start<-NULL
para.est<-NULL
para.est1<-NULL
wsse<-NULL
wsse1<-NULL
now=proc.time()
for (k1 in seq(k.g)) {
  for (k2 in seq(k.g)) {
    for (k3 in seq(k.g)) {
      for (k4 in seq(k.g)) {
        for (k5 in seq(k.g)) {
          now1=proc.time()
          tmp=try(oneiteration.dm.ds.gml(X.last,0*pars+(c(k1,k2,1,1,k3,k4,1,k5)-1)/(k.g-1),Z,Zder))
          if (class(tmp)=="try-error"){
            #Broke first iteration
            conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
            conv.case[k1,k2,k3,k4,k5]<-0
          }
          else{
           if (min(tmp$newTHETA)<0){#Negative first iteration
            conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3] 
            conv.case[k1,k2,k3,k4,k5]<-(-1)
           }
           else{#Continue first iteration
            #Iteration 1
            para.now<-tmp$newTHETA
            wsse.now<-tmp$WSSE
            new.theta<-c(tmp$newTHETA[1:2],0,0,tmp$newTHETA[3:4],0,tmp$newTHETA[5])
            names(new.theta)[c(3,4,7)]=c("dm","ds","gml")
            tmp1=try(oneiteration.dm.ds.gml(X.last,new.theta,Z,Zder))
            if (class(tmp1)=="try-error") {#broke second iteration
             conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
             conv.case[k1,k2,k3,k4,k5]<-(-2)               
             para.est<-rbind(para.est,para.now)
             wsse<-c(wsse,wsse.now)
             conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
             time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
             para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5)-1)/(k.g-1))
            }
            else{
             if (min(tmp1$newTHETA)<0) {#Neg the second iteration
              conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
              conv.case[k1,k2,k3,k4,k5]<-(-2)               
              para.est<-rbind(para.est,para.now)
              wsse<-c(wsse,wsse.now)
              conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
              time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
              para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5)-1)/(k.g-1))
             }
             else{#Continue iterations
               ## iterate more times
               ni=1
               while (ni <niter) {
                 para.now<-tmp1$newTHETA
                 wsse.now<-tmp1$WSSE
                 new.theta<-c(para.now[1:2],0,0,para.now[3:4],0,para.now[5])
                 names(new.theta)[c(3,4,7)]=c("dm","ds","gml")
                 tmp1=try(oneiteration.dm.ds.gml(tmp1$newX,new.theta,Z,Zder)) 
                 if (class(tmp1)=="try-error"){#broke ni+2 iteration
                   conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
                   conv.case[k1,k2,k3,k4,k5]<-(-ni-2)               
                   para.est<-rbind(para.est,para.now)
                   wsse<-c(wsse,wsse.now)
                   conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
                   time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
                   para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5)-1)/(k.g-1))
                   ni=niter #stop the loop
                 }
                 else{
                   if (min(tmp1$newTHETA)<0) {#Neg ni+2iteration
                    conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
                    conv.case[k1,k2,k3,k4,k5]<-(-ni-2)               
                    para.est<-rbind(para.est,para.now)
                    wsse<-c(wsse,wsse.now)
                    conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
                    time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
                    para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5)-1)/(k.g-1))
                    ni=niter #stop the loop
                   }
                   else{
                    if(tmp1$WSSE>wsse.now) {#Converge
                      conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
                      conv.case[k1,k2,k3,k4,k5]<-ni+2               
                      para.est<-rbind(para.est,para.now)
                      wsse<-c(wsse,wsse.now)
                      conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
                      time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
                      para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5) -1)/(k.g-1))
                      ni=niter #stop the loop
                    }
                    else {
                      if(ni==niter-1) {#Reach maximum iteration
                        conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
                        conv.case[k1,k2,k3,k4,k5]<-ni+2               
                        para.est<-rbind(para.est,tmp1$newTHETA)
                        wsse<-c(wsse,tmp1$WSSE)
                        conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
                        time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
                        para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5) -1)/(k.g-1))
                      }
                      # Continue iteration
                      ni=ni+1
                    }
                   }
                 }
               }# while loop
             }
            }
           }
          }
}}}}}
usedtime=proc.time()-now
usedtime

#########################
#### Summarize results
#########################
write.table(cbind(wsse,time1,conv1,para.est,para.start),file="resultsOCLP.txt")
mean(conv.time)
#[1] 0.7220988  #from new run on another machine?
#[1] 0.7785473
sum(wsse<113)
#[1] 46
#### Find best OCLP fit from a grid search of starting values
tmp.theta<-para.est[wsse==min(wsse),]
#          rm           rs           dl          gms          gsl 
#1.463910e-05 4.776081e-05 3.962229e+00 1.383559e-01 6.110069e-01 
para.start[wsse==min(wsse),]
#[1] 0.5 0.0 0.5 1.0 1.0
##### Compute RSS (log10 scale) for OCLP's fit
new.theta<-c(tmp.theta[1:2],0,0,tmp.theta[3:4],0,tmp.theta[5])
  names(new.theta)[3]="dm"
  names(new.theta)[4]="ds"
  names(new.theta)[7]="gml"
out.tmp     <- as.data.frame(ode(func = ODEmod, y = xini0,
                         parms = new.theta, times = fitteddata$time-5))
sum(((out.tmp[,2:4]-log(fitteddata[2:4]))/log(10))^2)
#[1] 18.49252

######################################################################
####### Do OCLP-NLS ##################################################
######################################################################
k.g<-3
conv.case<-array(NA,rep(k.g,5))
conv.time<-array(NA,rep(k.g,5))
conv1<-NULL
time1<-NULL
para.start<-NULL
para.est<-NULL
para.est1<-NULL
wsse<-NULL
rss<-NULL
now=proc.time()
for (k1 in seq(k.g)) {
  for (k2 in seq(k.g)) {
    for (k3 in seq(k.g)) {
      for (k4 in seq(k.g)) {
        for (k5 in seq(k.g)) {
          now1=proc.time()
          tmp=try(oneiteration.dm.ds.gml(X.last,0*pars+(c(k1,k2,1,1,k3,k4,1,k5)-1)/(k.g-1),Z,Zder))
          if (class(tmp)=="try-error"){
            #Broke first iteration
            conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
            conv.case[k1,k2,k3,k4,k5]<-0
          }
          else{
           if (min(tmp$newTHETA)<0){#Negative first iteration
            conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3] 
            conv.case[k1,k2,k3,k4,k5]<-(-1)
           }
           else{#Continue first iteration
            #Iteration 1
            new.theta<-c(tmp$newTHETA[1:2],0,0,tmp$newTHETA[3:4],0,tmp$newTHETA[5])
            names(new.theta)[c(3,4,7)]=c("dm","ds","gml")
            tmp1=try(oneiteration.dm.ds.gml(X.last,new.theta,Z,Zder))
            if (class(tmp1)=="try-error") {#broke second iteration
             conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
             conv.case[k1,k2,k3,k4,k5]<-(-2)               
            }
            else{
             if (min(tmp1$newTHETA)<0) {#Neg the second iteration
              conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
              conv.case[k1,k2,k3,k4,k5]<-(-2)               
             }
             else{                   ## Do NLS here
                   tmp=try(nls(
                         yfit ~ ODEsolver.0dsdmgml(x1,x2,x3, thepar), 
                         control=nls.control(warnOnly=T,tol=1e-4),
                         start=list(x1=xini[1],x2=xini[2],x3=xini[3], 
#                         thepar= para.est[70,])))
                         thepar=tmp1$newTHETA)))
                   conv.time[k1,k2,k3,k4,k5]<-(proc.time()-now1)[3]
                   if (class(tmp)=="try-error") {
                     conv.case[k1,k2,k3,k4,k5]<-(-4)
                   }
                   else {
                     conv.case[k1,k2,k3,k4,k5]<-4
                     para.est<-rbind(para.est,coef(tmp))
                     wsse<-rbind(wsse,tmp1$WSSE)
                     rss<-rbind(rss,sum((resid(tmp)/log(10))^2))
                     conv1<-c(conv1,conv.case[k1,k2,k3,k4,k5])
                     time1<-c(time1,conv.time[k1,k2,k3,k4,k5])
                     para.start<-rbind(para.start,(c(k1,k2,k3,k4,k5)-1)/(k.g-1))
                   } 
#                 }
#               }
             }
            }
           }
          }
}}}}}
usedtime=proc.time()-now
usedtime

write.table(cbind(rss,time1,conv1,wsse,para.est,para.start),file="resultsOCLP-NLS.txt")
mean(conv.time)
#[1] 7.294774
sum(rss<15.78)
#[1] 48
min(rss)
#[1] 15.76867